"""Control the settings of the project.

Allows reading from a settings file and writing to it.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar, TypeVar, cast

import rtoml

from simple_toml_settings.exceptions import (
    SettingsMutuallyExclusiveError,
    SettingsNotFoundError,
    SettingsSchemaError,
)
from simple_toml_settings.xdg_config import xdg_config_home

T = TypeVar("T", bound="TOMLSettings")


@dataclass
class TOMLSettings:
    """The main settings class.

    The only required argument is the app_name, which is used to create the
    settings folder. The settings_folder and settings_file_name are optional and
    will default to the app_name preceded by a '.' and config.toml
    respectively.
    """

    app_name: str
    settings_file_name: str = "config.toml"
    auto_create: bool = True
    local_file: bool = False
    flat_config: bool = False
    xdg_config: bool = False
    allow_missing_file: bool = False

    # the schema_version is used to track changes to the settings file.
    schema_version: str = "none"

    _instances: ClassVar[dict[type[TOMLSettings], TOMLSettings]] = {}

    _ignored_attrs: set[str] = field(
        default_factory=lambda: {
            "_instances",
            "app_name",
            "settings_folder",
            "settings_file_name",
            "auto_create",
            "local_file",
            "flat_config",
            "xdg_config",
            "allow_missing_file",
        }
    )

    _mutually_exclusive: set[str] = field(
        default_factory=lambda: {"local_file", "flat_config", "xdg_config"}
    )

    def __post_init__(self) -> None:
        """Create the settings folder if it doesn't exist."""
        self.settings_folder = self.get_settings_folder()

        # if we allow a missing file, we don't want to auto-create it
        if self.allow_missing_file:
            self.auto_create = False

        # ensure only one of the mutually exclusive options is set
        check_exclusive = {
            attr for attr in self._mutually_exclusive if getattr(self, attr)
        }
        if len(check_exclusive) > 1:
            raise SettingsMutuallyExclusiveError(attrs=check_exclusive)

        # load (or create) the settings file
        self.load()

    def get_settings_folder(self) -> Path:
        """Return the settings folder. If it doesn't exist, create it.

        Take into account the `local_file` and `flat_config` settings.
        Note that `local_file` takes precedence over `flat_config`.
        """
        if self.local_file:
            return Path.cwd()

        if self.flat_config:
            return Path.home()

        settings_folder: Path = Path.home() / f".{self.app_name}"

        if self.xdg_config:
            settings_folder = xdg_config_home() / f"{self.app_name}"
        if not settings_folder.exists():
            settings_folder.mkdir(parents=True)

        return settings_folder

    def __post_create_hook__(self) -> None:
        """Allow further customization after a new settings file is created.

        It is provided so that you can add any additional settings that you
        might need, or get information from the user. The subclass should
        override this method, by default it does nothing.

        The save() method IS called after we run this automatically, it should
        never be called manually.
        """

    @classmethod
    def get_instance(
        cls: type[T],
        app_name: str,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> T:
        """Class method to get or create the Settings instance.

        This is optional (and experimental), and is provided to allow for a
        singleton pattern in derived classes. It is not required to use this
        class.
        """
        if cls not in cls._instances:
            cls._instances[cls] = cls(app_name, *args, **kwargs)
        return cast(T, cls._instances[cls])

    def get_attrs(self, *, include_none: bool = False) -> dict[str, Any]:
        """Return a dictionary of our setting values.

        Values that are None are EXCLUDED by default, but can be included by
        setting 'include_none' to True.
        """
        return {
            a: getattr(self, a)
            for a in dir(self)
            if not a.startswith("_")
            and a not in self._ignored_attrs
            and not callable(getattr(self, a))
            and (include_none or getattr(self, a) is not None)
        }

    def save(self) -> None:
        """Save the settings to the settings file."""
        rtoml.dump(
            {self.app_name: self.get_attrs()},
            self.settings_folder / self.settings_file_name,
        )

    def load(self) -> None:
        """Load the settings from the settings file."""
        try:
            settings = rtoml.load(
                self.settings_folder / self.settings_file_name
            )
        except FileNotFoundError as exc:
            if self.auto_create:
                self.__post_create_hook__()
                self.save()
            elif self.allow_missing_file:
                return
            else:
                message = "Cant find a Config File, please create one."
                raise SettingsNotFoundError(message) from exc
            return

        # Check if 'schema_version' is present and matches the required one
        file_schema_version = str(
            settings[self.app_name].get("schema_version", None)
        )
        if file_schema_version.lower() not in {self.schema_version, "none"}:
            raise SettingsSchemaError(
                expected=self.schema_version, found=file_schema_version
            )
        for key, value in settings[self.app_name].items():
            setattr(self, key, value)

    def get(
        self,
        key: str,
    ) -> Any:  # noqa: ANN401
        """Get a setting by key."""
        try:
            return getattr(self, key)
        except AttributeError:
            return None

    def set(
        self,
        key: str,
        value: Any,  # noqa: ANN401
        *,
        autosave: bool = True,
    ) -> None:
        """Set a setting by key and value.

        If autosave is True (the default), the settings will be saved to the
        settings file each time it is called.
        """
        setattr(self, key, value)
        if autosave:
            self.save()

    def list_settings(self) -> dict[str, Any]:
        """Return a dictionary of settings."""
        return self.get_attrs()



"""Test the settings module."""

from pathlib import Path

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from simple_toml_settings.exceptions import (
    SettingsMutuallyExclusiveError,
    SettingsNotFoundError,
    SettingsSchemaError,
)
from simple_toml_settings.settings import TOMLSettings
from simple_toml_settings.xdg_config import xdg_config_home

from .conftest import SettingsExample


class CustomSettings(TOMLSettings):
    """Skeleton Class for testing."""

    my_var: bool = False

    def __post_create_hook__(self) -> None:
        """Override the post create hook."""


class TestSettings:
    """Contains tests for the settings module."""

    TEST_APP_NAME = "test_app"
    SETTINGS_FILE_NAME = "config.toml"
    SETTINGS_FILE_CONTENT = """
[test_app]
test_string_var = 'local_app'
schema_version= '1'
"""

    def test_config_file_auto_created(self, settings: SettingsExample) -> None:
        """Test that the settings file is created if it doesn't exist."""
        assert settings.settings_folder.exists()
        assert settings.settings_folder.is_dir()
        assert settings.settings_folder.name == f".{self.TEST_APP_NAME}"
        assert settings.settings_file_name == self.SETTINGS_FILE_NAME

    def test_exception_raised_on_missing_config_if_auto_create_is_false(
        self, fs: FakeFilesystem
    ) -> None:
        """Test that the settings file is not created if auto_create False."""
        fs.create_dir(Path.home())

        with pytest.raises(SettingsNotFoundError):
            TOMLSettings("test_app", auto_create=False)

    def test_allow_missing_file_disables_auto_create(
        self, fs: FakeFilesystem
    ) -> None:
        """Test that allow_missing_file disables auto_create."""
        fs.create_dir(Path.home())
        settings = CustomSettings(
            "test_app", auto_create=True, allow_missing_file=True
        )
        assert settings.settings_folder.exists()
        assert settings.settings_folder.is_dir()
        assert settings.settings_folder.name == f".{self.TEST_APP_NAME}"
        # assert that the config file is not created
        assert not (settings.settings_folder / self.SETTINGS_FILE_NAME).exists()
        # assert that 'auto_create' is False
        assert not settings.auto_create

    def test_no_exception_raised_on_missing_config_if_allow_no_file_is_true(
        self, fs: FakeFilesystem
    ) -> None:
        """Test the 'allow_missing_file' option.

        If 'allow_missing_file' is True, and 'auto_create' is False, then the
        settings file is not created, and no exception is raised.
        """
        fs.create_dir(Path.home())

        settings = CustomSettings(
            "test_app", auto_create=False, allow_missing_file=True
        )
        assert settings.settings_folder.exists()
        assert settings.settings_folder.is_dir()
        assert settings.settings_folder.name == f".{self.TEST_APP_NAME}"
        # assert that the config file is not created
        assert not (settings.settings_folder / self.SETTINGS_FILE_NAME).exists()

        # make sure we can still get values from the settings object
        assert settings.get("app_name") == "test_app"
        assert settings.get("my_var") is False

    def test_local_config(self, fs: FakeFilesystem) -> None:
        """Test that local_config loads settings from the local directory."""
        fs.create_file(
            self.SETTINGS_FILE_NAME,
            contents=self.SETTINGS_FILE_CONTENT,
        )
        settings = TOMLSettings("test_app", local_file=True, schema_version="1")
        assert settings.get("app_name") == "test_app"
        assert settings.get("test_string_var") == "local_app"

    def test_flat_config(self, flat_settings: SettingsExample) -> None:
        """Test that flat_config loads settings from home folder directly."""
        assert flat_settings.get("app_name") == "test_app"
        assert flat_settings.get("test_string_var") == "test_value"

        assert flat_settings.settings_folder == Path.home()
        assert Path(Path.home() / self.SETTINGS_FILE_NAME).exists()

    def test_xdg_config(self, xdg_settings: SettingsExample) -> None:
        """Test that settings file is created in the xdg_config_home folder."""
        assert xdg_settings.settings_folder.exists()
        assert xdg_settings.settings_folder.is_dir()
        assert xdg_settings.settings_folder.name == f"{self.TEST_APP_NAME}"
        assert xdg_settings.settings_file_name == self.SETTINGS_FILE_NAME
        assert (
            xdg_settings.settings_folder
            == xdg_config_home() / f"{self.TEST_APP_NAME}"
        )

        assert xdg_settings.get("app_name") == "test_app"
        assert xdg_settings.get("test_string_var") == "test_value"

    def test_settings_from_environment(self, monkeypatch, fs) -> None:
        """Test that the settings file is read from the xdg variable."""
        monkeypatch.setenv("XDG_CONFIG_HOME", "/path/validity/matters/not")

        settings = SettingsExample("test_app", xdg_config=True)
        assert (
            settings.settings_folder
            == Path("/path/validity/matters/not") / "test_app"
        )

    def test_post_create_hook_is_called(
        self, fs: FakeFilesystem, mocker: MockerFixture
    ) -> None:
        """Test that the post_create_hook is called after file created."""
        fs.create_dir(Path.home())

        mocker.patch.object(TOMLSettings, "__post_create_hook__")
        settings = TOMLSettings("test_app")

        assert settings.__post_create_hook__.called

    def test_post_create_hook_is_called_for_custom_class(
        self, fs: FakeFilesystem, mocker
    ) -> None:
        """Test that the post_create_hook is called after file created."""
        fs.create_dir(Path.home())

        mocker.patch.object(CustomSettings, "__post_create_hook__")
        settings = CustomSettings("test_app")

        assert settings.__post_create_hook__.called

    def test_get_attrs(self, settings) -> None:
        """Test that we can get the attributes of the settings object."""
        attrs = settings.get_attrs()
        assert attrs["schema_version"] == "none"
        assert attrs["test_string_var"] == SettingsExample.test_string_var
        assert attrs["test_int_var"] == SettingsExample.test_int_var

    def test_get_attrs_ignores_none_by_default(self, settings) -> None:
        """Test that None attributes are not returned."""
        settings.set("new_key", None)
        attrs = settings.get_attrs()
        assert "new_key" not in attrs

    def test_save_ignores_none_values(self, settings) -> None:
        """Test that None attributes are not saved."""
        settings.set("new_key", None)
        settings.save()
        attrs = settings.get_attrs()
        assert "new_key" not in attrs

    def test_get_attrs_returns_none_if_include_none_true(
        self, settings
    ) -> None:
        """Test that None attributes are returned if required is True."""
        settings.set("new_key", None)
        attrs = settings.get_attrs(include_none=True)
        assert attrs["new_key"] is None

    def test_get(self, settings) -> None:
        """Test we can get settings."""
        assert settings.get("app_name") == self.TEST_APP_NAME
        assert (
            settings.get("test_string_var") == SettingsExample.test_string_var
        )
        assert settings.get("test_int_var") == SettingsExample.test_int_var

    def test_get_missing_setting(self, settings: SettingsExample) -> None:
        """Test that 'None' is returned when a setting is missing."""
        assert settings.get("missing_setting") is None

    def test_set(self, settings: SettingsExample) -> None:
        """Test that a setting can be set."""
        settings.set("app_name", "new_test_app")
        assert settings.get("app_name") == "new_test_app"

    def test_add_and_list_setting(self, settings: SettingsExample) -> None:
        """Add a new setting and list all settings."""
        settings.set("new_key", "new_value")
        settings_dict = settings.list_settings()
        assert settings_dict["new_key"] == "new_value"

    def test_add_none_value(self, settings: SettingsExample) -> None:
        """Test that a setting can be set to None."""
        settings.set("new_key", None)
        assert settings.get("new_key") is None

    def test_load_settings(self, settings: SettingsExample) -> None:
        """Test that settings are loaded from the settings file."""
        settings.load()
        # length is 3 because of the 'schema' setting
        assert len(settings.list_settings()) == 3  # noqa: PLR2004

        assert settings.get("schema_version") == "none"
        assert settings.list_settings()["test_string_var"] == "test_value"
        assert (
            settings.list_settings()["test_int_var"]
            == SettingsExample.test_int_var
        )

    def test_set_schema_version(self, settings: SettingsExample) -> None:
        """Test that the schema_version can be set using the 'set' method."""
        settings.set("schema_version", "1.0.0")
        settings.load()
        assert settings.get("schema_version") == "1.0.0"

    def test_autosave(self, settings: SettingsExample) -> None:
        """Test that settings are auto saved when autosave is True."""
        settings.set("test_string_var", "new_value")
        settings.load()
        assert settings.get("test_string_var") == "new_value"

    def test_no_autosave(self, settings: SettingsExample) -> None:
        """Test that settings are not auto saved when autosave is False."""
        settings.set("test_string_var", "new_value", autosave=False)
        settings.load()
        assert settings.get("test_string_var") == "test_value"

    def test_custom_file_name(self, fs: FakeFilesystem) -> None:
        """Test that the settings file name can be customized."""
        custom_file_name = "custom_config.toml"

        fs.create_dir(Path.home())
        settings = TOMLSettings("test_app", settings_file_name=custom_file_name)

        assert (Path.home() / ".test_app" / custom_file_name).exists()

        assert settings.settings_file_name == custom_file_name
        assert settings.settings_folder.name == f".{self.TEST_APP_NAME}"
        assert settings.settings_folder / settings.settings_file_name

    def test_items_on_ignored_attrs(self, settings: SettingsExample) -> None:
        """Test that the ignored attributes are not returned by items()."""
        list_settings = settings.list_settings()

        for setting in settings._ignored_attrs:  # noqa: SLF001
            assert setting not in list_settings

    def test_schema_version_mismatch_raises_error(
        self, fs: FakeFilesystem
    ) -> None:
        """Test that a schema version mismatch raises SettingsSchemaError."""
        fs.create_file(
            self.SETTINGS_FILE_NAME,
            contents=self.SETTINGS_FILE_CONTENT,
        )
        with pytest.raises(SettingsSchemaError):
            TOMLSettings("test_app", local_file=True, schema_version="2")

    @pytest.mark.parametrize("value", ["none", "NONE", "None"])
    def test_none_schema_does_not_raise_error(
        self, fs: FakeFilesystem, value: str
    ) -> None:
        """Test that a 'none' schema does NOT raise SettingsSchemaError."""
        fs.create_file(
            self.SETTINGS_FILE_NAME,
            contents="[test_app]\ntest_var = 'schema test'\n"
            f"schema_version='{value}'",
        )

        # this should NOT raise an exception
        TOMLSettings("test_app", local_file=True, schema_version="2")

    def test_missing_schema_does_not_raise_error(
        self, fs: FakeFilesystem
    ) -> None:
        """Test that a 'none' schema does NOT raise SettingsSchemaError."""
        fs.create_file(
            self.SETTINGS_FILE_NAME,
            contents="[test_app]\ntest_var = 'schema test'\n",
        )

        # this should NOT raise an exception
        TOMLSettings("test_app", local_file=True, schema_version="2")

    def test_get_instance(self, fs: FakeFilesystem) -> None:
        """Test that we can get the instance of the settings object."""
        fs.create_dir(Path.home())
        assert isinstance(TOMLSettings.get_instance("test_app"), TOMLSettings)

    def test_get_instance_is_singleton(self, fs: FakeFilesystem) -> None:
        """Test that the instance is a singleton."""
        fs.create_dir(Path.home())
        instance1 = TOMLSettings.get_instance("test_app")
        instance2 = TOMLSettings.get_instance("test_app")
        assert instance1 is instance2

    def test_get_instance_with_custom_class_is_singleton(
        self, fs: FakeFilesystem
    ) -> None:
        """Test that the instance is a singleton."""
        fs.create_dir(Path.home())
        instance1 = CustomSettings.get_instance("test_app")
        instance2 = CustomSettings.get_instance("test_app")
        assert instance1 is instance2

    def test_get_instance_with_multiple_subclasses_not_equal(
        self, fs: FakeFilesystem
    ) -> None:
        """Test that two subclasses are different instances."""
        fs.create_dir(Path.home())
        instance1 = CustomSettings.get_instance("test_app")
        instance2 = SettingsExample.get_instance("test_app")

        assert instance1 is not instance2  # type: ignore[comparison-overlap]

    def test_get_instance_with_custom_class(self, fs: FakeFilesystem) -> None:
        """Test that we can get the instance of a custom settings class."""
        fs.create_dir(Path.home())
        assert isinstance(
            CustomSettings.get_instance("test_app"), CustomSettings
        )

    def test_get_instance_attribute(self, fs: FakeFilesystem) -> None:
        """Test that we can get the instance of the settings object."""
        fs.create_dir(Path.home())
        settings = CustomSettings.get_instance("test_app")
        assert settings.my_var is False

    def test_mutually_exclusive_attributes(self, fs: FakeFilesystem) -> None:
        """Test that mutually exclusive attributes are handled."""
        fs.create_dir(Path.home())

        error_pattern = (
            r"Only one of (flat_config|local_file), "
            r"(flat_config|local_file) can be True\."
        )

        with pytest.raises(SettingsMutuallyExclusiveError, match=error_pattern):
            CustomSettings("test_app", local_file=True, flat_config=True)



"""Define exceptions for the simple_toml_settings package."""


class SettingsError(Exception):
    """Base exception for settings errors."""


class SettingsNotFoundError(SettingsError):
    """The Settings file has not been found.

    This will be raised if the settings file is not found and auto_create is
    False.
    """


class SettingsSchemaError(SettingsError):
    """The settings file schema does not match the required level."""

    def __init__(self, expected: str, found: str) -> None:
        """Define a custom response for this Exception."""
        self.expected = expected
        self.found = found
        super().__init__(
            f"Schema version mismatch: Expected {self.expected}, "
            f"found {self.found} in file."
        )


class SettingsMutuallyExclusiveError(SettingsError):
    """Two or more mutually exclusive settings are set to True."""

    def __init__(self, attrs: set[str]) -> None:
        """Define a custom response for this Exception."""
        self.attrs = attrs
        super().__init__(f"Only one of {', '.join(self.attrs)} can be True.")


# temporary alias for backwards compatibility
SettingsNotFound = SettingsNotFoundError
